import sys, requests, tarfile, io, zipfile
from contextlib import closing


def debug(msg): pass # print('## ' + msg)

missing_deps = { 'mock' : [ 'unittest2' ],
                 'Whoosh' : [ 'unittest2', 'nose'],  # 'pypkgs.pytest
                 'nose' : [ 'unittest2'],  # paste?
                 'pylint' : [ 'unittest2', 'logilab-astng'],
                 'gunicorn' : [ 'unittest2', ],  #'pypkgs.pytest' ],  # 'pypkgs.argparse'
                 'lxml' : [ 'pkgs.libxslt', 'pkgs.libxml2', 'unittest2' ], # pkgs.pytest
                 'suds' : [ 'python-dateutil' ],
                 #'paste' : [ 'pypkgs.pytest' ],
                 #'South' : [ 'pypkgs.pytest' ],
                 #'Django' : [ 'pypkgs.pytest' ],
                 'django_tastypie' : [ 'python-dateutil', 'mock' ],
                 'django_nose' : [ 'South' ],
                 'setproctitle' : [ ], # 'pypkgs.pytest' ],
                 'MySQL_python' : [ 'pkgs.mysql', 'pkgs.zlib' ],
                 'docutils' : [ 'logilab-astng' ],
                 'logilab_common' : [ 'unittest2' ],
}


package_extras = { 'MySQL_python': """
              preConfigure = ''
                sed -i '/use_setuptools/d' setup.py
              '';""",
               }


def name_fix(name):
    if name == 'webob': return 'WebOb'
    if name == 'whoosh': return 'Whoosh'
    if name == 'BeautifulSoup4': return 'beautifulsoup4'
    return name


def fpartition(pstr, pfunc):
    for p1 in range(len(pstr)):
        if pfunc(pstr[p1]):
            break
    if p1 < len(pstr) - 1:
        for p2 in range(p1, len(pstr)):
            if not pfunc(pstr[p2]):
                return pstr[:p1], pstr[p1:p2], pstr[p2:]
        return pstr[:p1], pstr[p1:], ''
    return pstr,'',''


class Version:
    def __init__(self, vspec):
        self._ver = vspec.split('.')
    @property
    def major(self): return self._ver[0]
    @property
    def minor(self): return self._ver[1] if len(self._ver) > 1 else 0
    @property
    def bugfix(self): return self._ver[2] if len(self._ver) > 2 else 0
    def __str__(self): return '.'.join(self._ver)
    def __eq__(self, o): return self.major == o.major and self.minor == o.minor
    def __ne__(self, o): return self.major != o.major or self.minor != o.minor
    def __ge__(self, o): return self.major > o.major or (self.major == o.major and self.minor >= o.minor)
    def __le__(self, o): return self.major < o.major or (self.major == o.major and self.minor <= o.minor)
    def __gt__(self, o): return self.major > o.major or (self.major == o.major and self.minor > o.minor)
    def __lt__(self, o): return self.major < o.major or (self.major == o.major and self.minor < o.minor)


class Relation:
    def __init__(self, relver=None):
        if relver and relver != ('',''):
            self._rel = relver[0]
            if not relver[1]:
                raise ValueError('Invalid version specification: %s'%str(relver))
            self._ver = Version(relver[1])
        else:
            self._rel = self._ver = None
    def __str__(self): return '*' if self._rel is None else (self._rel + ' ' +
                                                             str(self._ver))
    @property
    def relation(self): return self._rel
    @property
    def version(self): return self._ver
    def upper_limit(self):
        if self._rel in [None, '>=', '>']: return None
        return self._ver
    def merge(self, other, throwMsg = None):
        if self._rel is None: return other
        if other._rel is None: return self
        if self._rel == other._rel and self._ver == other._ver: return self
        if self._rel == '==':
            if other._rel == '>=' and self._ver >= other._ver: return self
            if other._rel == '<=' and self._ver <= other._ver: return self
            if other._rel == '>' and self._ver > other._ver: return self
            if other._rel == '<' and self._ver < other._ver: return self
        elif self._rel == '>=':
            if other._rel == '==' and other._ver >= self._ver: return other
            if other._rel == '>=':
                if other._ver >= self._ver: return other
                return self
            if other._rel == '>':
                if other._ver > self._ver: return other
                return self
            # if other._rel == '<=' -- no bounded-range support yet
        elif self._rel == '<=':
            if other._rel == '==' and other._ver <= self._ver: return other
            if other._rel == '<=':
                if other._ver <= self._ver: return other
                return self
            if other._rel == '<':
                if other._ver < self._ver: return other
                return self
            # if other._rel == '>=' -- no bounded-range support yet
        elif self._rel == '>':
            if other._rel == '==' and other._ver > self._ver: return other
            if other._rel == '>=':
                if other._ver >= self._ver: return other
                return self
            if other._rel == '>':
                if other.ver > self._ver: return other
                return self
        elif self._rel == '<':
            if other._rel == '==' and other._ver < self._ver: return other
            if other._rel == '<=':
                if other._ver <= self._ver: return other
                return self
            if other._rel == '<':
                if other._ver < self._ver: return other
                return self
        if throwMsg:
            raise ValueError('Version relation (%s) incompatible with (%s); %s'%(
                str(self), str(other), throwMsg))
        return None


class Packages:
    def __init__(self):
        self._pkgs = {}  # key = pkg.name, value = pkg
    def add(self, pkg):
        if pkg.name in self._pkgs:
            self._pkgs[pkg.name] = self._pkgs[pkg.name].merge(pkg, ' in existing packages')
        else:
            self._pkgs[pkg.name] = pkg
    def append(self, pkg): self.add(pkg)
    def __iter__(self):
        for pname in self._pkgs:
            yield self._pkgs[pname]
    def __contains__(self, o):
        if isinstance(o, PkgInfo): return o.name in self._pkgs
        return o in self._pkgs
    def __getitem__(self, key): return self._pkgs[key]
    def __setitem__(self, key, val): self._pkgs[key] = val
    def __len__(self): return len(self._pkgs)
    def copy(self):
        newP = Packages()
        newP._pkgs = self._pkgs.copy()
        return newP


class PkgInfo:
    def __init__(self, name, pypi_json, relver=None, primary=False):
        self._name   = name
        self._relver = relver
        self._json   = pypi_json
        self._cachedreqs = None
        self._primary = primary
        debug('json %s'%str(pypi_json))
        print('# name %s'%name)
        debug('keys %s'%self._json.keys())
        for each in self._json.get('urls', []):
            debug(each.get('packagetype', 'unknown-package'))
            # has_sig upload_time comment_text python_version url md5_digest downloads filename packagetype size
            if each.get('packagetype', 'unknown') != 'sdist':
                continue
            if getattr(self, '_url', None):
                if '.tar' in each['url'][-20:]:
                    if len(each['url']) > len(self._url):
                        self._url = each['url']
                        self._md5 = each['md5_digest']
            else:
                self._url = each['url']
                self._md5 = each['md5_digest']
        if 'download_url' in self._json.get('info', {}) and \
           not getattr(self, '_url', '') and \
           self._json['info']['download_url'] != 'UNKNOWN':
            self._url = self._json['info']['download_url']
        self._version = self._json.get('info', {}).get('version', '')
        for key in self._json.get('info', []):
            # maintainer docs_url requires_python maintainer_email cheesecake_code_kwalitee_id
            # keywords package_url author author_email download_url platform version
            # cheesecake_documentation_id description release_url downloads classifiers
            # name bugtrack_url license summary home_page stable_version
            # cheesecake_installability_id
            if key in 'requires_python package_url download_url platform version name home_page release_url'.split():
                debug('...%s ==> %s'%(key, self._json['info'][key]))
    @property
    def name(self): return self._name
    @property
    def version(self): return self._version
    @property
    def relver(self): return self._relver
    @property
    def url(self): return getattr(self, '_url', 'none')
    @property
    def md5(self): return getattr(self, '_md5', '')
    @property
    def varname(self): return self.name.replace('-', '_')
    @property
    def isPrimary(self): return self._primary
    def has_source(self): return self.url and self.md5
    def requires(self):
        "Returns the set of packages required by this one, in requirements.txt format"
        if not self.url: return {}
        if self._cachedreqs is None:
            self._cachedreqs = self._requires()
            # Now do some repairs for some packages that seem to have missing dependencies
            if self.varname in missing_deps:
                for mdep in missing_deps[self.varname]:
                    if mdep not in self._cachedreqs:
                        self._cachedreqs[mdep] = Relation()
        return self._cachedreqs
    def _requirestar(self):
        subreqs = {}
        with closing(requests.get(self.url, stream=True)) as rtf:
            rtfcontent = io.BytesIO(rtf.content)
            with closing(tarfile.open(fileobj=rtfcontent, mode='r|*')) as tarf:
                for entry in tarf:
                    if 'requires.txt' in entry.name:
                        with closing(tarf.extractfile(entry)) as reqf:
                            subreqs.update(parse_reqfile(reqf))
        return subreqs
    def _requireszip(self):
        subreqs = {}
        with closing(requests.get(self.url, stream=True)) as rtf:
            rtfcontent = io.BytesIO(rtf.content)
            with closing(zipfile.ZipFile(rtfcontent, 'r')) as zipf:
                for entry in zipf.infolist():
                    if 'requires.txt' in entry.filename:
                        with closing(zipf.open(entry)) as reqf:
                            subreqs.update(parse_reqfile(reqf))
        return subreqs
    def _requires(self):
        if self.url.endswith('.tar'): return self._requirestar()
        if self.url.endswith('.tbz2'): return self._requirestar()
        if self.url.endswith('.tgz'): return self._requirestar()
        if self.url.endswith('.tar.gz'): return self._requirestar()
        if self.url.endswith('.zip'): return self._requireszip()
        return {}
    def merge(self, other, throwMsg = None):
        if self.name != other.name:
            if throwMsg:
                raise ValueError('Packages have different names ("%s" and "%s")'%(
                    self.name, other.name))
            return None
        self._relver = self._relver.merge(other.relver,
                                          ('Package ' + self.name + ' ' + throwMsg)
                                          if throwMsg else None)
        if self._relver is None: return None
        if not self._cachedreqs and other._cachedreqs:
            self._cachedreqs = other._cachedreqs
        return self
    def attributes(self):
        homepage = self._json.get('info', {}).get('home_page', '')
        if not homepage or homepage == 'UNKNOWN':
            homepage = 'https://pypi.python.org/pypi/' + self.name
        return { 'name': self.name, 'version': self.version, 'url': self.url,
                 'varname': self.varname, 'md5': self.md5,
                 'homepage': homepage,
                 'summary': self._json.get('info', {}).get('summary', ''),
             }


def fetchURL(name, verspec=None):
    # Get the PyPi information about the specified package, honoring
    # most version specifications.  This will, however, ignore a
    # version specification of >= or >, because the query from PyPi
    # will automatically get the latest version, which is always
    # expected to satisfy >= and > requirements.  In addition, getting
    # the exact version for a >= can be problematic (e.g. Sphynx
    # requires Jinja2, which requires Babel>=0.8, but Babel 0.8's pypi
    # information is rubbish and will fail)
    ulim = verspec.upper_limit()
    if ulim:
        if verspec.relation == '<': raise NotImplemented('no < limit retrieval')
        return 'http://pypi.python.org/pypi/%s/%s/json'%(name, str(verspec.version))
    return 'http://pypi.python.org/pypi/%s/json'%(name)


def pypi_json(url):
    try:
        return requests.get(url).json()
    except ValueError:
        return {}


def parse_reqfile(rf):
    reqs = {}
    for line in rf:
        ll = line.split('#')[0].strip()
        if not ll:
            continue
        name,rel,ver = [x.strip() for x in fpartition(ll, lambda c: c in '=<>')]

        # Skip known bad packages and cases we can't handle
        if '+' in name: continue
        if name[0] == '[': continue  # sections like [tests], [doc], etc.
        if name == 'setuptools': continue
        name = name_fix(name)

        try:
            reqs[name] = Relation((rel,ver))
        except ValueError:
            sys.stderr.write('*** Error parsing requirement: "%s"\n'%ll)
            raise
    return reqs

def read_reqfile(reqfile):
    with open(reqfile, 'r') as rf:
        return parse_reqfile(rf)


class NixFile:
    def __init__(self, envName):
        self.ename = envName
        self.fname = envName + 'Env.nix'
    def write(self, pkgs, pyversions):
        pversions = pyversions.split(',')
        with open(self.fname, 'w') as nf:
            self._write_instructions(nf, pversions)
            self._write_hdr(nf)
            nf.write('let\n\n')
            for pkg in pkgs:
                self._write_pkg(nf, pkg, pkgs)
            self._write_python_env_decl(nf, pkgs)
            for pver in pversions:
                self._write_python_version_env(nf, pver)
            nf.write('in\n')
            self._write_footer(nf, pversions)
        return self.fname

    def _write_instructions(self, of, pversions):
        of.write('''# GENERATED Nix operations (via nixpip)
# Usage: $ nix-env -f %(fname)s -i env-%(pyver)s%(name)s
#        $ load-env-%(pyver)s%(name)s
#        %(pyver)s%(name)s$ cmds...
#        %(pyver)s%(name)s$ exit
#        $
#
# Supported environments: '''%
                 { 'name': self.ename,
                   'fname': self.fname,
                   'pyver': pversions[0]})
        of.write(', '.join(['env-%s%s'%(v, self.ename) for v in pversions]))
        of.write('\n\n')

    def _write_hdr(self, of):
        of.write('''
with import <nixpkgs> {};

''')

    def _write_pkg(self, of, pkg, pkgs):
        srcvalidation = ('  md5 = "%s"; '%pkg.md5) if pkg.md5 else ''
        if pkg.has_source():
            of.write('''
  %(varname)s = python: pypkgs: pypkgs.buildPythonPackage rec {
              name = "%(name)s-%(version)s";
              doCheck = false;
              src = pkgs.fetchurl {
                      url = "%(url)s";
                    %%s};'''%pkg.attributes()%srcvalidation)
            if pkg.requires():
                of.write('''
              buildInputs = [\n''');
                for req in pkg.requires():
                    if req.startswith('pypkgs.') or req.startswith('pkgs.'):
                        of.write(' ' * 22 + req + '\n')
                    elif pkgs[req].has_source():
                        of.write(' ' * 22 + '(' + pkgs[req].varname + ' python pypkgs)\n')
                    else:
                        print('No source for %s (%s/%s), so just using pypkgs reference.'%(
                            req, pkgs[req].name, pkgs[req].varname))
                        of.write(' ' * 22 + 'pypkgs.' + pkgs[req].varname + '\n')
                of.write(' '*42 + '];')
            of.write(package_extras.get(pkg.varname, ''))
            of.write('''
              meta = { homepage = %(homepage)s;
                       description = "%(summary)s";
                     };
                    };
'''%pkg.attributes())
        else:
            print('# Warning: %(name)s has no download info'
                  ', falling back to: %(varname)s' % pkg.attributes())

    def _write_python_env_decl(self, of, pkgs):
        of.write('''  
  declPyEnv = name: pyfull: pypkgs: rec {

               pkglist = map (x: x pyfull.python pypkgs) [\n''')
        for pkg in pkgs:
            if pkg.has_source() and pkg.isPrimary:
                # Declared above
                of.write(' '*28 + pkg.varname + '\n')
        of.write(' '*28 + '] ++ [' +
                 ' '.join(['pypkgs.'+x.varname
                           for x in pkgs if not x.has_source() and x.isPrimary
                           and not x.varname.startswith('pypkgs.')
                           and not x.varname.startswith('pkgs.')]) +
                 ' '.join([x.varname for x in pkgs if not x.has_source() and x.isPrimary
                           and x.varname.startswith('pypkgs.') and x.varname.startswith('pkgs.')]) +
                 '''];

               penv = myEnvFun {
                        name = name;
                        shell = "${pkgs.bashInteractive}/bin/bash";
                        buildInputs = pkglist ++ [ pyfull stdenv ];
                        extraCmds = ''
                          export PYTHONHOME=${pyfull}
                          export PYTHONPATH=${pkgs.lib.concatStringsSep ":" (map (x: "${x}/lib/${pyfull.libPrefix}/site-packages") pkglist)}:$PYTHONPATH
                          export LOADED_ENV=${name}
                          unset http_proxy
                                    ''; };
  };

''')

    def _write_python_version_env(self, of, pver):
        pfullver = pver.replace('py', 'python')
        name = self.ename
        of.write('  %(pver)senv = declPyEnv "%(pver)s%(name)s" '
                 '%(pfullver)s %(pfullver)sPackages;\n'%locals())

    def _write_footer(self, of, pyvers):
        of.write('  [ ' + ' '.join(['%senv'%v for v in pyvers]) + ' ]')
        of.write('\n\n')


def getPkgInfo(name, relver, primary=False):
    # Fix up some names....bcause pypi is case-sensitive, but apparently pip is not.
    return PkgInfo(name_fix(name), pypi_json(fetchURL(name, relver)), relver,
                   primary=primary)

def add_dependencies(pkgs, checkpkgs):
    newpkgs = Packages()  # don't change pkgs while we are iterating through it
    for pkg in checkpkgs:
        sys.stdout.write('Dependencies for %s... '%pkg.name)
        sys.stdout.flush()
        newreqs = pkg.requires()
        sys.stdout.write(str(dict([(n,str(r)) for n,r in newreqs.items()])))
        sys.stdout.write('\n')
        for newr in newreqs:
            npkg = getPkgInfo(newr, newreqs[newr])
            if newr not in pkgs and newr not in newpkgs: newpkgs.add(npkg)
            elif newr in pkgs: pkgs[newr] = pkgs[newr].merge(npkg, "new dependency")
            elif newr in newpkgs: newpkgs[newr] = newpkgs[newr].merge(npkg, "new dependencies")
    for newpkg in newpkgs:
        pkgs.add(newpkg)
    return pkgs, newpkgs


def nixpip(reqfile, name='pip', pyversions='py26,py27,py33'):
    debug('Process requirements file: %s'%reqfile)
    reqs = read_reqfile(reqfile)
    pkgs = Packages()
    for k,v in reqs.items():
        pkgs.add(getPkgInfo(k, v, True))

    newpkgs = pkgs.copy()
    for ii in range(20):
        pkgs, newpkgs = add_dependencies(pkgs, newpkgs)
        if not newpkgs:
            break

    if newpkgs:
        raise RuntimeError('Failed to converge on stable dependency set.'
                           '  Current set: %s'
                           ',  latest additions: %s'%(str([p.name for p in pkgs]),
                                                      str([p.name for p in newpkgs])))

    for pkg in pkgs:
        print ('  %-12s  %s  %s'%(pkg.name, str(pkg.relver), pkg.url))

    nixfile = NixFile(name)
    print(nixfile.write(pkgs, pyversions))
            

if __name__ == "__main__":
    if len(sys.argv) not in [3,4]:
        print('Usage: nixpip {pip-requirements-file} [envname] [python-version-list]')
        print('''

The envname is the desired name for the environment to be created from
the specified pip requirements file (defaulting to "pip").  The
optional python-version-list is a comma-separated list of python
versions to generate environments for, specified similarly to tox
(defaulting to:  py26,py27,py33) ''')
        sys.exit(2)
    nixpip(sys.argv[1],
           'pip' if len(sys.argv) < 3 else sys.argv[2],
           'py26,py27,py33' if len(sys.argv) < 4 else sys.argv[3])
